// Doulos Verification TechNote 1
// file Rcvr.sv
//
// This model detects and reports activity on the
// Send_Rcv interface, which has an 8-bit data line
// and two handshake signals.  It also checks for
// some protocol errors.  In the examples, it is
// used to confirm that the testbench is correctly
// generating stimulus.

module Rcvr (input [7:0] D, input logic Valid, output logic Ready);

  timeunit 1ns; timeprecision 1ns;

  // Seed for random time delay generation
  //
  int seed = 1234;

  // Detect and report data transactions.
  // Introduce random time delays in the handshake.
  //
  initial begin

    // At startup, clear down the Ready response and
    // wait for the Valid signal to settle at an idle level.
    Ready <= 0;
    wait (Valid === 1'b0);

    // Loop on detecting and reporting data.
    forever begin

      // Wait for Valid to go true.
      @(posedge Valid);
      // Random time delay before responding
      #($dist_uniform(seed, 3, 12));
      // Handshake response...
      Ready <= 1;
      // ... accept and report the data
      $display("%m at time %0d: transaction D='h%h", $time, D);

      // Wait for the transaction to end...
      @(negedge Valid);
      // random delay before ending response
      #($dist_uniform(seed, 3, 12));
      Ready <= 0;

      // then loop around to wait for the next.
    end
  end

  // Some protocol error checking.
  initial begin
    // Wait until everything is quiet at the start
    wait (Valid === 1'b0);
    // Launch the checking loops.
    fork
      // Check #1: behavior of Valid
      forever @(Valid) begin
        // 1a: Valid should never be X/Z
        assert (!$isunknown(Valid)) else $error("X or Z on Valid");
        // 1b: Valid rise must occur when Ready is false;
        //     Valid fall must occur when Ready is true
        assert (Valid != Ready) else $error("Handshake error");
      end
      // Check #2: behavior of data
      forever @(D)
        // 1a: Data must not change while Valid is true
        assert (!Valid) else $error("D unstable");
    join
  end

endmodule
