// Doulos Verification TechNote 1
// file TechNote1_pkg.sv
//
// This file contains the Data_Generator and
// Sender_BFM classes needed for the testbench
// described in the TechNote.  It also contains
// a wrapper class to allow the virtual interface
// to be configured, an interface definition to
// provide the necessary connections, and an
// environment class that instances
// Data_Generator and Sender_BFM together.


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

// This interface contains all the signals needed
// to support the Send_Rcv connection.
//
interface SendRcvIntf;

  timeunit 1ns; timeprecision 1ns;

  logic Ready, Valid;
  logic [7:0] D;

endinterface

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

`include "ovm_macros.svh"

package TechNote1_pkg;

  timeunit 1ns; timeprecision 1ns;

  import ovm_pkg::*;

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // This class defines a wrapper for a virtual
  // interface reference, making it easy to
  // configure other components' virtual interfaces.
  //
  class SendRcvIntf_wrapper extends ovm_object;
    virtual SendRcvIntf pins;
    function new(virtual SendRcvIntf where);
      super.new("SendRcvIntf_wrapper");
      pins = where;
    endfunction : new
  endclass : SendRcvIntf_wrapper

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // This is the Sender_BFM class as described in TechNote 1.

  class Sender_BFM extends ovm_component;

    // A blocking_put_imp allows this object to expose
    // its put() method.
    //
    ovm_blocking_put_imp #(byte, Sender_BFM) send_export;

    // Virtual interface connection
    SendRcvIntf_wrapper pins_wrapper;
    protected virtual SendRcvIntf pins;

    function new(string name, ovm_component parent = null);
      super.new(name, parent);
    endfunction : new

    // In the build() phase, we create any internal structures
    // of this component, including its blocking_put_imp.
    //
    function void build();
      super.build();
      send_export = new("send_export", this);
      assert (pins_wrapper != null)
        else $error("pins_wrapper not configured in %s", get_full_name());
      pins = pins_wrapper.pins;
    endfunction : build

    // The blocking_put_imp exposes this object's put() method.
    // The exported method MUST be named put().
    //
    task put(byte D);
      // In practice the call will probably be forwarded directly
      // to other methods that do the real work.
      send(D);
    endtask : put

    task run();
      pins.Valid <= 0;
    endtask : run

    local task send(byte D);
      pins.D <= D;
      #5
      pins.Valid <= 1;
      wait (pins.Ready);
      #5 pins.Valid <= 0;
      wait (!pins.Ready);
    endtask : send

    `ovm_component_utils_begin(Sender_BFM)
      `ovm_field_object(pins_wrapper, OVM_ALL_ON)
    `ovm_component_utils_end
  endclass : Sender_BFM

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // This is the Data_Generator class as described in TechNote 1.

  class Data_Generator extends ovm_component;

    // This component has a blocking_put_port so that it can call
    // the put() method exported by some other component.
    ovm_blocking_put_port #(byte) send_port;

    function new(string name, ovm_component parent = null);
      super.new(name, parent);
    endfunction : new

    function void build();
      super.build();
      send_port = new("send_port", this);
    endfunction : build

    rand byte data;

    task run();
      repeat (100) begin
        void'(randomize(data));
        // To give away the data we just created, we call
        // our port's put() method.  It's a blocking method
        // so it will not return immediately.  We neither
        // know nor care where the put() method is implemented.
        send_port.put(data);
      end
    endtask

    `ovm_component_utils_begin(Data_Generator)
      `ovm_field_int(data, OVM_ALL_ON)
    `ovm_component_utils_end
  endclass : Data_Generator

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // This class implements the environment shown in
  // Figure 4 of TechNote 1.

  class TechNote1_Fig4_env extends ovm_env;

    // Instance variables for the generator and BFM
    Data_Generator gen;
    Sender_BFM     bfm;

    function new(string name, ovm_component parent = null);
      super.new(name, parent);
    endfunction : new

    // In the build() phase, construct the two sub-components.
    function void build();
      super.build();
      gen = Data_Generator::type_id::create("gen", this);
      bfm = Sender_BFM::type_id::create("bfm", this);
    endfunction : build

    // In the connect() phase, link sub-components' ports.
    function void connect();
      super.connect();
      gen.send_port.connect(bfm.send_export);
    endfunction : connect

    `ovm_component_utils(TechNote1_Fig4_env)
  endclass : TechNote1_Fig4_env

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // To illustrate connections up and down the component hierarchy,
  // these two classes act as wrappers for the Sender_BFM and
  // Data_Generator classes.  Iinstance names of the child
  // components in these classes match those in Figure 6 of
  // TechNote 1.

  class Sender_BFM_Wrapper extends ovm_component;

    Sender_BFM compD;
    ovm_blocking_put_export #(byte) send_export;

    function new(string name, ovm_component parent = null);
      super.new(name, parent);
    endfunction : new

    // In the build() phase, construct the sub-component.
    function void build();
      super.build();
      compD = Sender_BFM::type_id::create("compD", this);
      send_export = new("send_export", this);
    endfunction : build

    // In the connect() phase, link sub-component's export
    // to the matching export on this object.  As always,
    // the REQUIRING export's connect() method is called
    // with the PROVIDING export as its argument.

    function void connect();
      super.connect();
      this.send_export.connect(compD.send_export);
    endfunction : connect

    `ovm_component_utils(Sender_BFM_Wrapper)
  endclass : Sender_BFM_Wrapper

  class Data_Generator_Wrapper extends ovm_component;

    Data_Generator compA;
    ovm_blocking_put_port #(byte) send_port;

    function new(string name, ovm_component parent = null);
      super.new(name, parent);
    endfunction : new

    // In the build() phase, construct the sub-component
    // and the port.
    function void build();
      super.build();
      compA = Data_Generator::type_id::create("compA", this);
      send_port = new("send_port", this);
    endfunction : build

    // In the connect() phase, link sub-component's port
    // to the matching port on this object.  As always,
    // the REQUIRING port's connect() method is called
    // with the PROVIDING port as its argument.

    function void connect();
      super.connect();
      compA.send_port.connect(this.send_port);
    endfunction : connect

    `ovm_component_utils(Data_Generator_Wrapper)
  endclass : Data_Generator_Wrapper

  //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // This class implements the environment shown in
  // Figure 6 of TechNote 1.  The same instance names
  // have been used as in that Figure.

  class TechNote1_Fig6_env extends ovm_env;

    // Instance variables for the generator and BFM
    Data_Generator_Wrapper compB;
    Sender_BFM_Wrapper     compC;

    function new(string name, ovm_component parent = null);
      super.new(name, parent);
    endfunction : new

    // In the build() phase, construct the two sub-components.
    function void build();
      super.build();
      compB = Data_Generator_Wrapper::type_id::create("compB", this);
      compC = Sender_BFM_Wrapper::type_id::create("compC", this);
    endfunction : build

    // In the connect() phase, link sub-components' ports.
    function void connect();
      super.connect();
      compB.send_port.connect(compC.send_export);
    endfunction : connect

    `ovm_component_utils(TechNote1_Fig6_env)
  endclass : TechNote1_Fig6_env

endpackage

